/* vi:set sm ai

	blinded input routine
				by ann@mrit.mei.co.jp

*/

#include "cba.h"

extern int trace;
extern int term;
extern writemsg();
extern int waitfd();
extern int getcar();

char
blindchar()
{
    char c;
    static int toggle = 0;
    static char bstr1[] = "aiueo";
    static char bstr2[] = "kstnhmrjvbw";

    c = toggle?
	bstr1[rand() % (sizeof(bstr1) - 1)]:
	bstr2[rand() % (sizeof(bstr2) - 1)];
    toggle = 1 - toggle;

    return(c);
}

char *
blindread(fi, fo, mesg, blind)
    int fi, fo;
    char *mesg;
    int  blind;
{
    char c, *cp;
    struct termios org, new;
    struct timeval to;

    static char cbuf[MAXFLINE + 1];

    do
    {
	int rv;

	lsleep(1);
	write(fo, mesg, strlen(mesg));
	write(fo, " : ", 3);
	tcdrain(fo);

	tcgetattr(fi, &org);
	tcgetattr(fi, &new);
	
	new.c_lflag &= ~ECHO;	/* no echo */
	new.c_lflag &= ~ICANON;	/* raw mode */
	new.c_cflag &= ~CLOCAL;	/* watch modem control lines */

	/* discard characters can be read */
	new.c_cc[VMIN]  = 1;
	new.c_cc[VTIME] = 0;

	tcsetattr(fi, TCSANOW, &new);

	to.tv_sec =
	to.tv_usec = 0;

	while(waitfd(fi, &to) > 0)
	{
	    rv = read(fi, &c, 1);

	    if (!(trace & TRACE_CTRL))
	    	continue;

	    if (rv < 0)
		writemsg(MAINT, "blindread error: '%s'\n", strerror(errno));

	    else if (rv == 1)
	    {
	    	if (c >= ' ' && !(c & 0x80))
		    writemsg(TRACE, "blindread ignores '%c'\n", c);
		else
		    writemsg(TRACE, "blindread ignores '\\%03o'\n", c & 0xff);
	    }

	}

	/* */
	new.c_cc[VMIN]  = 1;
	new.c_cc[VTIME] = 0;

	tcsetattr(fi, TCSANOW, &new);

	cp = cbuf; 

	do
	{
	    errno = 0;

	    rv = waitfd(fi, NULL);
	    if (rv < 0 || (rv = read(fi, &c, 1)) < 0)
	    {
	    	/* interrupted or error occured */
	    	if (errno && (trace & TRACE_CTRL))
		    writemsg(TRACE, "blindread error: %s\n", strerror(errno));

	    	return NULL;
	    }


	    if (!term && !getcar(fi))
	    	return NULL;

	    if (cp == cbuf + MAXFLINE)
		break;

	    switch(c)
	    {
	        char x;

	    case '\r':
	    case '\n':
		c = '\n';
		*cp = NUL;
		if (cp == cbuf)
		    break;

		while(cp < cbuf + 10)
		{
		    x = blindchar();
		    write(fo, &x, 1);
		    tcdrain(fo);
		    cp++;
		}
		break;

	    case '\010':	/* ^H  */
	    case '\177':	/* DEL */
		if (cp != cbuf)
		{
		    cp--;
		    write(fo, "\010 \010", 3);
		    tcdrain(fo);
		}
		continue;

	    case '\025':	/* ^U */
		for(;cp != cbuf; cp--)
		{
		    write(fo, "\010 \010", 3);
		    tcdrain(fo);
		}
		continue;

	    case '\004':	/* ^D */
	    	return NULL;

	    case '\377':	/* PPP */
	    	if (cbuf[0] = '~')
		{
		    *(cp++) = c;
		    *cp = NUL;
		    break;
		}

	    default:
		*(cp++) = c;
		x = blind ? blindchar(): c;
		write(fo, &x, 1);
		tcdrain(fo);
	    }

	} while(c != '\n');

	write(fo, "\n", 1);
	tcdrain(fo);

	tcsetattr(fi, TCSANOW, &org);

    } while (!cbuf[0]);

    return cbuf;
}

