/* vi:set sm ai

	setup file reader for CBA

				by ann@mrit.mei.co.jp
*/

#include "cba.h"

extern int trace;	/* trace level */
extern writemsg();

static lineesc();

/*
 *	read device(tty) configuration file 
 */

int
readconf(fname, tty, syscp, devcp, mtypp)
    char *fname;
    char *tty;
    struct sys *syscp;
    struct dev *devcp;
    struct mtyp *mtypp;
{
    FILE *fp;
    int devfound;
    int modemfound;
    int typcount;
    int i;

    devfound = 0;
    modemfound = 0;
    typcount = 0;

    syscp->verify = NULL;
    syscp->to_ac  = 30;
    syscp->to_in  = 60;
    syscp->to_back= 60;
    syscp->monint = 30;		/* min.  */
    syscp->maxuse = 1440;	/* 24 hour */
    syscp->maxwait= 1440;	/* 24 hour */
    devcp->name   = "/dev/tty";
    devcp->baud   = "9600";
    syscp->nserv1 =
    syscp->nserv2 = NULL;
    devcp->modem.name  = "[modem]";
    devcp->modem.setup = "(setup)";
    devcp->modem.dial  = "(dial)";
    devcp->modem.hook  = "(hook)";
    devcp->modem.reset = "(reset)";
    devcp->modem.hsync = "(hsync)";
    devcp->modem.hang  =
    devcp->modem.esc   = NULL;
    for(i = 0; i < MAXTRACE; i++)
	devcp->modem.trace[i].type = TT_EMPTY;
    devcp->dext   = "";
    devcp->ip     =
    devcp->peer   = "localhost";

    fp = fopen(fname, "r");
    if (!fp)
    {
	writemsg(SYSTM, "no device configuration file '%s'.\n", fname);
	return -1;
    }
    while(!feof(fp))
    {
    	static char *parsetios();
	char cbuf[MAXFLINE];
	char p1[MAXFLINE];
	char com[MAXFLINE];

	cbuf[0] = NUL;
	fgets(cbuf, MAXFLINE, fp);
	lineesc(cbuf);

	com[0] = NUL;
	if (!sscanf(cbuf,"%s", com))
	    continue;

	if (!strcmp(com, "VERIFY"))
	{
	    static char verify[20];

	    p1[0] = NUL;
	    sscanf(cbuf + strlen(com), "%s", p1);
	    strncpy(verify, p1, 20);
	    verify[19] = NUL;
	    syscp->verify = verify;

	    if (trace & TRACE_FILE)
		writemsg(TRACE, "(VERIFY) %s\n", syscp->verify);
	    continue;
	}

	if (!strcmp(com, "ACC_TIMEOUT"))
	{
	    p1[0] = NUL;
	    sscanf(cbuf + strlen(com), "%s", p1);
	    syscp->to_ac = atoi(p1);

	    if (trace & TRACE_FILE)
		writemsg(TRACE, "(Acc. Ver. TIMEOUT) %d sec.\n", syscp->to_ac);
	    continue;
	}

	if (!strcmp(com, "TIMEOUT") || !strcmp(com, "IN_TIMEOUT"))
	{
	    p1[0] = NUL;
	    sscanf(cbuf + strlen(com), "%s", p1);
	    syscp->to_in = atoi(p1);

	    if (trace & TRACE_FILE)
		writemsg(TRACE, "(Call-in TIMEOUT) %d sec.\n", syscp->to_in);
	    continue;
	}

	if (!strcmp(com, "BACK_TIMEOUT") || !strcmp(com, "OUT_TIMEOUT"))
	{
	    p1[0] = NUL;
	    sscanf(cbuf + strlen(com), "%s", p1);
	    syscp->to_back = atoi(p1);

	    if (trace & TRACE_FILE)
		writemsg(TRACE, "(Call-back TIMEOUT) %d sec.\n",
							    syscp->to_back);
	    continue;
	}

	if (!strcmp(com, "BACK_INTERVAL"))
	{
	    p1[0] = NUL;
	    sscanf(cbuf + strlen(com), "%s", p1);
	    syscp->intvl = atoi(p1);

	    if (trace & TRACE_FILE)
		writemsg(TRACE, "(Call-back INTERVAL) %d sec.\n", syscp->intvl);
	    continue;
	}

	if (!strcmp(com, "BACK_RETRY"))
	{
	    p1[0] = NUL;
	    sscanf(cbuf + strlen(com), "%s", p1);
	    syscp->retry = atoi(p1);

	    if (trace & TRACE_FILE)
		writemsg(TRACE, "(Call-back RETRY) %d times.\n", syscp->retry);
	    continue;
	}

	if (!strcmp(com, "MONITOR"))
	{
	    p1[0] = NUL;
	    sscanf(cbuf + strlen(com), "%s", p1);
	    syscp->monint = atoi(p1);

	    if (trace & TRACE_FILE)
		writemsg(TRACE, "(Monitor Interval) %d min.\n", syscp->monint);
	    continue;
	}

	if (!strcmp(com, "MAX_USE"))
	{
	    p1[0] = NUL;
	    sscanf(cbuf + strlen(com), "%s", p1);
	    syscp->maxuse = atoi(p1);

	    if (trace & TRACE_FILE)
		writemsg(TRACE, "(Session TIMEOUT) %d min.\n", syscp->maxuse);
	    continue;
	}

	if (!strcmp(com, "MAX_WAIT"))
	{
	    p1[0] = NUL;
	    sscanf(cbuf + strlen(com), "%s", p1);
	    syscp->maxwait = atoi(p1);

	    if (trace & TRACE_FILE)
		writemsg(TRACE, "(Callin TIMEOUT) %d min.\n", syscp->maxwait);
	    continue;
	}

	if (!strcmp(com, "MS_DNS"))
	{
	    static char ns1buf[128];
	    static char ns2buf[128];
	    char *nb;

	    p1[0] = NUL;
	    sscanf(cbuf + strlen(com), "%s", p1);
	    nb = syscp->nserv1 ? ns1buf: ns2buf;

	    strncpy(nb, p1, 128);
	    *(nb + 127) = NUL;
	    if (trace & TRACE_FILE)
		writemsg(TRACE, "(MS_DNS) %s\n", nb);

	    (syscp->nserv1 ? syscp->nserv2: syscp->nserv1) = nb;
	    continue;
	}

	if (!strcmp(com, "PROTO") || !strcmp(com, "TYPE"))
	{
	    char p2[MAXFLINE], p3[MAXFLINE], p4[MAXFLINE];

	    p1[0] = p2[0] = p3[0] = p4[0] = NUL;
	    sscanf(cbuf + strlen(com), "%s %s %s %s", p1, p2, p3, p4);

	    if (strcmp(p2, "*") && strcmp(p2, devcp->modem.name))
	    {
		if (trace & TRACE_FILE)
		    writemsg(TRACE, "(PROTO) %s ignored.\n", p1);
	    	continue;
	    }

	    if (trace & TRACE_CTRL)
		writemsg(TRACE, "(PROTO) %s:%s found.\n", p1, p2);

	    if (typcount >= MAXMTYPES)
	    {
		writemsg(SYSTM, "(PROTO:%1) too many PROTO entry.\n", p1);
	    	continue;
	    }

	    strncpy(mtypp->name,   p1, sizeof(mtypp->name) - 1);
	    strncpy(mtypp->setup,  p3, sizeof(mtypp->setup)- 1);
	    strncpy(mtypp->dial,   p4, sizeof(mtypp->dial) - 1);

	    mtypp->name[ sizeof(mtypp->name) - 1] =
	    mtypp->setup[sizeof(mtypp->setup)- 1] =
	    mtypp->dial[ sizeof(mtypp->dial) - 1] = NUL;

	    if (trace & TRACE_FILE)
		writemsg(TRACE, "(PROTO) %s: %s\n",
			mtypp->name, mtypp->setup);

	    typcount ++;
	    (++ mtypp)->name[0] = NUL;

	    continue;

	}

	if (!strcmp(com, "TTY") || !strcmp(com, "DEV"))
	{
	    char p2[MAXFLINE], p3[MAXFLINE], p4[MAXFLINE],
	         p5[MAXFLINE], p6[MAXFLINE];
	    static char name[80], modem[40], ip[40], peer[40], dext[40];

	    p1[0] = p2[0] = p3[0] = p4[0] = p5[0] = p6[0] = NUL;
	    sscanf(cbuf + strlen(com),
	    			"%s %s %s %s %s %s", p1, p2, p3, p4, p5, p6);

	    if (strcmp(p1, tty))
	    {
		if (trace & TRACE_FILE)
		    writemsg(TRACE, "(TTY) %s ignored.\n", p1);
	    	continue;
	    }

	    devfound = 1;

	    strncpy(name,   p1, sizeof(name) - 1);
	    strncpy(modem,  p3, sizeof(modem)- 1);
	    strncpy(ip,     p4, sizeof(ip)   - 1);
	    strncpy(peer,   p5, sizeof(peer) - 1);
	    strncpy(dext,   p6, sizeof(dext) - 1);

	    name[ sizeof(name) - 1] =
	    modem[sizeof(modem)- 1] =
	    ip[   sizeof(ip)   - 1] =
	    peer[ sizeof(peer) - 1] =
	    dext[ sizeof(dext) - 1] = NUL;

	    devcp->name  = name;
	    devcp->baud  = parsetios(name, &(devcp->tio), p2);
	    devcp->modem.name = modem;
	    devcp->ip    = ip;
	    devcp->peer  = peer;
	    devcp->dext  = dext;

	    if (trace & (TRACE_FILE | TRACE_MODEM))
	    {
		writemsg(TRACE, "(TTY) terminal %s  %s\n", devcp->name, p2);
		writemsg(TRACE, "(TTY) modem: %s (ext. '%s')\n",
					devcp->modem.name, devcp->dext);
		writemsg(TRACE, "(TTY) ip: %s peer: %s\n",
					devcp->ip, devcp->peer);
	    }
	    continue;
	}

	if (!strcmp(com, "MODEM") || !strcmp(com, "DCE"))
	{
	    char p2[MAXFLINE], p3[MAXFLINE], p4[MAXFLINE],
	    	 p5[MAXFLINE], p6[MAXFLINE], p7[MAXFLINE];
	    static char dial[40], hook[20], reset[20], hsync[20],
	    		hang[20], esc[20];

	    p1[0] = p2[0] = p3[0] = p4[0] = p5[0] = p6[0] = p7[0] = NUL;
	    sscanf(cbuf + strlen(com), "%s %s %s %s %s %s %s",
						p1, p2, p3, p4, p5, p6, p7);
	    if (strcmp(p1, devcp->modem.name))
	    {
		if (trace & TRACE_FILE)
		    writemsg(TRACE, "(MODEM) %s ignored.\n", p1);
	    	continue;
	    }

	    if (trace & TRACE_CTRL)
		writemsg(TRACE, "(MODEM) %s found.\n", p1);

	    modemfound++;

	    strncpy(dial,   p2, sizeof(dial) - 1);
	    strncpy(hook,   p3, sizeof(hook) - 1);
	    strncpy(reset,  p4, sizeof(reset)- 1);
	    strncpy(hsync,  p5, sizeof(hsync)- 1);
	    strncpy(hang,   p6, sizeof(hang) - 1);
	    strncpy(esc,    p7, sizeof(esc)  - 1);

	    dial[ sizeof(dial) - 1] =
	    hook[ sizeof(hook) - 1] =
	    reset[sizeof(reset)- 1] =
	    hsync[sizeof(hsync)- 1] =
	    hang[ sizeof(hang) - 1] =
	    esc[  sizeof(esc)  - 1] = NUL;

	    devcp->modem.dial  = dial;
	    devcp->modem.hook  = hook;
	    devcp->modem.reset = reset;
	    devcp->modem.hsync = hsync;
	    devcp->modem.hang  = *hang ? hang: NULL;
	    devcp->modem.esc   = *esc  ? esc : NULL;

	    devcp->modem.hang  = NULL;
	    if (!*hang || !strcmp(hang, "BREAK"))
	    	devcp->modem.hang_method = HUP_BREAK;
	    else if (!strcmp(hang, "DTR"))
	    	devcp->modem.hang_method = HUP_DTR;
	    else
	    {
		devcp->modem.hang_method = HUP_EMPTY;
	        devcp->modem.hang  = hang;
	    }

	    if (trace & TRACE_FILE)
	    {
		writemsg(TRACE, "(MODEM) modem %s\n", devcp->modem.name);
		writemsg(TRACE, "(MODEM) reset: %s hsync: %s hang: %s\n",
				devcp->modem.reset,
				devcp->modem.hsync,
				devcp->modem.hang_method == HUP_BREAK
				  ? "BREAK"
				  : devcp->modem.hang_method == HUP_DTR
				    ? "DTR"
				    : devcp->modem.hang_method == HUP_EMPTY
				      ? devcp->modem.hang
				      : "?");
		writemsg(TRACE, "(MODEM) dial: %s hook:%s\n",
					devcp->modem.dial,
					devcp->modem.hook);
	    }
	    continue;
	}

	if (!strcmp(com, "SETUP") || !strcmp(com, "MINIT"))
	{
	    char p2[MAXFLINE];
	    static char setup[80];

	    p1[0] = p2[0] = NUL;
	    sscanf(cbuf + strlen(com), "%s %s", p1, p2);
	    if (strcmp(p1, devcp->modem.name))
	    {
		if (trace & TRACE_FILE)
		    writemsg(TRACE, "(SETUP) %s ignored.\n", p1);
	    	continue;
	    }
	    if (trace & TRACE_CTRL)
		writemsg(TRACE, "(SETUP) %s found.\n", p1);

	    modemfound++;

	    strncpy(setup,  p2, sizeof(setup)- 1);
	    setup[sizeof(setup)- 1] = NUL;
	    devcp->modem.setup = setup;

	    if (trace & TRACE_FILE)
	    {
		writemsg(TRACE, "(SETUP) modem %s\n", devcp->modem.name);
		writemsg(TRACE, "(SETUP) setup: %s\n", devcp->modem.setup);
	    }
	    continue;
	}


	if (!strcmp(com, "TRACE_INIT")
	 || !strcmp(com, "TRACE_RING")
	 || !strcmp(com, "TRACE_BACK")
	 || !strcmp(com, "TRACE_RTRY")
	 || !strcmp(com, "TRACE_DISC"))
	{
	    char p2[MAXFLINE];
	    char p3[MAXFLINE];
	    static char pbuf[1024], *cbp = pbuf, *cbe = pbuf + 1024;
	    char *cp, *cm, *nm;

	    p1[0] = p2[0] = p3[0] = NUL;
	    sscanf(cbuf + strlen(com), "%s %s %s", p1, p2, p3);
	    if (strcmp(p1, devcp->modem.name))
	    {
		if (trace & TRACE_FILE)
		    writemsg(TRACE, "(%s) %s ignored.\n", com, p1);
	    	continue;
	    }
	    if (trace & TRACE_CTRL)
		writemsg(TRACE, "(%s) %s found.\n", com, p1);

	    for(i = 0; i < MAXTRACE; i++)
		if (devcp->modem.trace[i].type == TT_EMPTY)
		    break;

	    if (i >= MAXTRACE)
	    {
		writemsg(MAINT, "(%s) %s : Too many trace command(%s).\n",
								com, p1, p2);
	    	continue;
	    }

	    for(cp = p2, cm = cbp; cbp < cbe && *cp; cbp++, cp++)
	    	*cbp = *cp;
	    *cbp++ = NUL;
	    for(cp = p3, nm = cbp; cbp < cbe && *cp; cbp++, cp++)
	    	*cbp = *cp;
	    *cbp++ = NUL;

	    if (cbp >= cbe)
	    {
		writemsg(MAINT, "(%s) %s : Too long trace strings(%s).\n",
								com, p1, p2);
	    	continue;
	    }

	    if      (!strcmp(com, "TRACE_INIT"))
		devcp->modem.trace[i].type = TT_INIT;
	    else if (!strcmp(com, "TRACE_RING"))
		devcp->modem.trace[i].type = TT_RING;
	    else if (!strcmp(com, "TRACE_BACK"))
		devcp->modem.trace[i].type = TT_BACK;
	    else if (!strcmp(com, "TRACE_RTRY"))
		devcp->modem.trace[i].type = TT_RTRY;
	    else if (!strcmp(com, "TRACE_DISC"))
		devcp->modem.trace[i].type = TT_DISC;

	    devcp->modem.trace[i].modem = cm;
	    devcp->modem.trace[i].name  = nm;

	    if (trace & (TRACE_CTRL | TRACE_FILE))
		writemsg(TRACE, "(%s %d) %s:0x%2x: %s\n", com, i,
						devcp->modem.trace[i].name,
						devcp->modem.trace[i].type,
						devcp->modem.trace[i].modem);
	    continue;
	}


	if (com[0])
	{
	    writemsg(SYSTM, "error: '%s'\n", cbuf);
	    continue;
	}
    }

    fclose(fp);
    return (devfound && modemfound >= 2 )? 1 : 0;
}

/*
 *	read phone number file 
 */

int
readphones(fname, phones, nphone)
    char  *fname;
    struct user phones[];
    int   nphone;
{
    FILE *fp;
    int  i;
    struct user *pp;

    fp = fopen(fname, "r");
    if (!fp)
    {
	writemsg(SYSTM, "no phone number file '%s'.\n", fname);
	return -1;
    }

    i = 0;
    pp = phones;
    *(pp->name) = NUL;

    while(!feof(fp) && i++ < nphone)
    {
	char cbuf[MAXFLINE];
	char name[MAXFLINE],
	     num[MAXFLINE],
	     id[MAXFLINE],
	     setup[MAXFLINE],
	     pname[MAXFLINE],
	     appli[MAXFLINE],
	     term[MAXFLINE];

	fgets(cbuf, MAXFLINE, fp);
	lineesc(cbuf);

	if (!cbuf[0])
	    continue;

	name[0] = num[0] = id[0] = setup[0] = pname[0]= appli[0] = NUL;
	strcpy(term, "vt100");

	sscanf(cbuf, "%s %s %s %s %s %s %s",
			name, num, id, setup, pname, appli, term);

	if (!name[0])
	    continue;

	name[ MAXNAME] =
	num[  MAXPHONE] =
	id[   MAXPHONEID] =
	setup[MAXSETUP] =
	appli[MAXAKEY] =
	term[ MAXTERM] = NUL;

	{
	    int len;
	    char *cp;

	    len = 0;
	    cp = pname;
	    while(*cp)
	    {
	    	int w;

		w = (*cp & 0x80)? 2 : 1;

		if (len + w >= MAXPNAME)
		{
		    *cp = NUL;
		    break;
		}
		cp  += w;
		len += w;
	    }
	}

	strcpy(pp->name,  name);
	strcpy(pp->phone, num);
	strcpy(pp->pid,   id);
	strcpy(pp->where, pname);
	strcpy(pp->setup, setup);
	strcpy(pp->appli, appli);
	strcpy(pp->term,  term);


	if (trace & TRACE_FILE)
	   writemsg(TRACE, "%8s [%6s]%12s<%s:%s>(%s) %s\n",
	       pp->name, pp->pid, pp->where, 
	       pp->phone, pp->appli,
	       pp->setup, pp->term ? pp->term: "unknown");

	*((++pp)->name) = NUL;
    }

    fclose(fp);
    return 0;
}

/*
 *	read application file 
 */
int
readappli(fname, akey, aargv, aasize, env)
    char  *fname;
    char  *akey;
    char  *aargv[];
    int   aasize;
    struct vars *env;
{
    FILE *fp;
    int  found;
    int  suid;

    void expandenv();

    fp = fopen(fname, "r");
    if (!fp)
    {
	writemsg(SYSTM, "no application file '%s'.\n", fname);
	return -1;
    }

    found = 0;
    suid = 0;
    while(!feof(fp))
    {
	static char cbuf[MAXFLINE];
	char abuf[MAXFLINE],
	     *ctop,
	     *cp, *ap;
	int  i, l;
	extern char *osid;

	fgets(cbuf, MAXFLINE, fp);

	/* skip preceeding spaces */
	for(cp = cbuf; *cp == ' ' || *cp == '\t'; cp++);

	/* skip comment line */
	if (*cp == '#' || *cp == NUL)
	    continue;

	/* skip unformatted line */
	for(cp = cbuf; *cp && *cp != ':' && *cp != '!'; cp++);
	if (!*cp)
	    continue;

	suid = (*cp != ':');	/* or '%' */
	*cp++ = NUL;
	ctop = cp;

	/* check application */
	cp = cbuf;
	if (!strncmp(osid, cp, strlen(osid)))
	    cp += strlen(osid);

	if (strcmp(cp, akey))
	{
	    if (trace & TRACE_FILE)
		writemsg(TRACE, "application '%s' ignored\n", cbuf);
	    continue;
	}

	if (trace & TRACE_FILE)
	    writemsg(TRACE, "application '%s' found\n", cbuf);

	/* expand environment */
	for(cp = ctop; *cp == ' ' || *cp == '\t'; cp++);
	expandenv(abuf, MAXFLINE, cp, env);

	cp = ctop;
	ap = abuf;
	while(*ap)
	    *cp++ = *ap++;
	*cp = NUL;

	if (trace & TRACE_FILE)
	    writemsg(TRACE, "expanded: %s\n", ctop);

	/* parse args */
	i = 0;
	cp = ctop;
	*aargv = NULL;

	while(*cp)
	{
	    while(*cp == ' ' || *cp == '\t' || *cp == '\n' || *cp == '\r')
	    	cp++;

	    *aargv = cp;
	    *++aargv = NULL;
	    if (++ i >= aasize)
	    {
		writemsg(SYSTM, "%s in file %s too long(>= %d).\n",
							akey, fname, i);
		fclose(fp);
		return -1;
	    }

	    while(*cp &&
	    	  *cp != ' ' && *cp != '\t' && *cp != '\n' && *cp != '\r')
	    	cp++;
	    *(cp ++) = NUL;
	}

	found = 1;
	break;
    }

    fclose(fp);

    if (!found)
    {
	writemsg(SYSTM, "application '%s' not found in '%s'.\n", akey, fname);
	return -1;
    }
    return suid ? 1: 0;
}


/*
 *	expand environment
 */

void
expandenv(dest, dlen, form, env)
    char *dest;
    int  dlen;
    char *form;
    struct vars *env;
{
    char *dp;
    char *cp;

    dp = dest;
    cp = form;

    while(*cp && dlen > 1)
	if (*cp == '%')
	{
	    struct vars *ep;
	    char *vp;

	    cp++;

	    for(ep = env; ep->name && ep->name != *cp; ep++);

	    if (ep->name)
	    {
		cp++;
		vp = ep->val;
		while(*vp && dlen > 0)
		{
		    *dp++ = *vp++;
		    dlen --;
		}
	    }
	    else
	    {
		*dp++ = '%';
		*dp++ = *cp++;
		dlen -= 2;
	    }
	}
	else
	{
	    *dp++ = *cp++;
	    dlen --;
	}

    *dp = NUL;
}


/*
 *	print application oriented message
 */
int
printamsg(fout, fname, akey, env)
    int fout;
    char  *fname;
    char  *akey;
    struct vars *env;
{
    FILE *fp;
    int  found;

    fp = fopen(fname, "r");
    if (!fp)
    {
	writemsg(SYSTM, "no application message file '%s'.\n", fname);
	return -1;
    }

    found = 0;

    while(!feof(fp))
    {
	char cbuf[MAXFLINE];
	char abuf[MAXFLINE],
	     *ctop,
	     *cp, *ap;
	int  l;
	int  wait;

	fgets(cbuf, MAXFLINE, fp);

	/* skip preceeding spaces */
	for(cp = cbuf; *cp == ' ' || *cp == '\t'; cp++);

	/* skip comment line */
	if (*cp == '#' || *cp == NUL)
	    continue;

	/* skip unformatted line */
	for(cp = cbuf; *cp && *cp != '{'; cp++);
	if (!*cp)
	    continue;

	*cp++ = NUL;
	ctop = cp;

	/* check application */
	for(cp = cbuf; *cp && *cp != '.'; cp++);
	if (*cp) /* == '.' */
	{
	    *cp++ = NUL;
	    wait = atoi(cp);
	}
	else
	    wait = 0;

	if (strcmp(cbuf, akey))
	{
	    if (trace & TRACE_FILE)
		writemsg(TRACE, "application message %s ignored\n", cbuf);

	    cp = ctop;
	    do
	    {
	    	while(*cp && *cp != '}')
		{
		    if (*cp == '%' && *(cp + 1))
		    	cp++;
		    cp++;
		}
		if (*cp)
		    break;

		fgets(cp = cbuf, MAXFLINE, fp);
	    } while(!feof(fp));

	    continue;
	}

	if (trace & TRACE_FILE)
	    writemsg(TRACE, "application message for %s found\n", cbuf);

	found = 1;

	if (wait)
	{
	    if (trace & TRACE_CTRL)
		writemsg(TRACE, "waiting %d sec.\n", wait);
	    sleep(wait);
	}

	cp = ctop;

	do
	{
	    int brk;

	    /* expand environment */
	    ap = abuf;
	    l = (int)strlen(cp);
	    brk = 0;
	    while(*cp && ap - abuf < MAXFLINE - l - 1)
	    {
	    	switch(*cp)
		{
		case '}':
		    brk = 1;
		    break;

		case '\n':
		case '\r':
		    cp++;
		    continue;

		case '%':
		    if (*(cp + 1))
		    {
			struct vars *ep;
			char *vp;

			cp++;

			for(ep = env; ep->name && ep->name != *cp; ep++);

			if (ep->name)
			{
			    cp++;
			    vp = ep->val;
			    while(*vp)
				*ap++ = *vp++;
			}
			else
			{
			    *ap++ = '%';
			    *ap++ = *cp++;
			}
			continue;
		    }

		default:
		    *ap++ = *cp++;
		    continue;
		}

		break;
	    }

	    *ap = NUL;

	    if (trace & TRACE_FILE)
		writemsg(TRACE, "message: '%s'\n", abuf);

	    write(fout, abuf, strlen(abuf));
	    if (!*cp)
		write(fout, "\r\n", 2);

	    if (brk)
	    	break;

	    fgets(cp = cbuf, MAXFLINE, fp);

	} while(!feof(fp));

	wait = 0;
    }

    fclose(fp);
    if (!found && (trace & TRACE_FILE))
    	writemsg(MAINT, "application message %s could not found.\n", akey);
    return found;
}

/*
 *	converting meta character and JIS to EUC
 */

static int
lineesc(line)
    char *line;
{
    char *cp, *cp2;
    char *dp;
    int kanji;

    kanji = 0;
    cp = dp = line;
    while(*cp)
    {
	switch(*cp)
	{
	case '\033':
	    if      (!strncmp(cp, "\033(B", 3) || 
		     !strncmp(cp, "\033(J", 3))
	    {
		kanji = 0;
		cp += 3;
		continue;
	    }
	    else if (!strncmp(cp, "\033$B", 3) || 
		     !strncmp(cp, "\033$@", 3))
	    {
		kanji = 1;
		cp += 3;
		continue;
	    }
	    cp++;
	    continue;

	case '#' :
	    if (kanji)
	    	break;

	case '\n':
	case '\r':
	    *dp++ = NUL;
	    cp++;
	    continue;
	    
	case '\\':
	    if (kanji)
	    	break;

	    cp2 = cp;
	    switch(*(++cp))
	    {
	    case '\\':*dp++ = '\\';   cp++; break;
	    case 'n': *dp++ = '\n';   cp++; break;
	    case 'r': *dp++ = '\r';   cp++; break;
	    case 't': *dp++ = '\t';   cp++; break;
	    case 'e': *dp++ = '\033'; cp++; break;
	    case '#': *dp++ = '#';    cp++; break;
	    default :
		if ('0' <= *cp && *cp <= '9')
		{
		    *dp    = (* cp++ - '0') << 6;
		    *dp   += (* cp++ - '0') << 3;
		    *dp++ +=  * cp++ - '0';
		}
	    }
	    continue;
	}

	*dp ++ = kanji? *cp ++ | 0x80: *cp ++;
    }
    return 0;
}

static struct name_val
{
    char *name;
    int  val;
} baudt[]=
    	{
	    {"9600",  B9600},
#ifdef B460800
	    {"460800", B460800},
#endif
#ifdef B307200
	    {"307200", B307200},
#endif
#ifdef B230400
	    {"230400", B230400},
#endif
#ifdef B153600
	    {"153600", B153600},
#endif
#ifdef B115200
	    {"115200", B115200},
#endif
#ifdef B76800
	    {"76800", B76800},
#endif
#ifdef B57600
	    {"57600", B57600},
#endif
#ifdef B38400
	    {"38400", B38400},
#endif
#ifdef B19200
	    {"19200", B19200},
#endif
#ifdef B4800
	    {"4800",  B4800},
#endif
#ifdef B2400
	    {"2400",  B2400},
#endif
#ifdef B1800
	    {"1800",  B1800},
#endif
#ifdef B1200
	    {"1200",  B1200},
#endif
#ifdef B600
	    {"600",   B600},
#endif
#ifdef B300
	    {"300",   B300},
#endif
#ifdef B200
	    {"200",   B200},
#endif
#ifdef B150
	    {"150",   B150},
#endif
#ifdef B134
	    {"134",   B134},
#endif
#ifdef B110
	    {"110",   B110},
#endif
#ifdef B75
	    {"75",   B75},
#endif
#ifdef B50
	    {"50",   B50},
#endif
	    {NULL,    0}
	},
	csizet[]={{"8",CS8},{"7",CS7}, {"6",CS6}, {"5", CS5},	{NULL, 0}},
	cstopt[]={{"1", 0}, {"2", CSTOPB},			{NULL, 0}},
	cpart[]= {{"N", 0}, {"O", PARENB|PARODD}, {"E", PARENB},{NULL, 0}},
	cflow[]= {{"N", 1}, {"H", 2}, {"X", 3},{NULL, 0}};

static char *
parsetios(tty, term, cpar)
    char *tty;
    struct termios *term;
    char *cpar;
{ 
    struct name_val *tp;
    char *ret = "";

    term->c_iflag =	0x0
			| IGNBRK	/**//* do not clear */
			| BRKINT	/**/
     			| IGNPAR	/**/
     		/*	| PARMRK	  */
     		/*	| INPCK		  */
     		/*	| ISTRIP	  */
     		/*	| INLCR		  */
     		/*	| IGNCR		  */
     			| ICRNL		/**/
     		/*	| IUCLC		  */
		/*	| IXON		  */
     		/*	| IXANY		  */
     		/*	| IXOFF		  */
     		/*	| IMAXBEL	  */
			;
    term->c_oflag =	0x0
			| OPOST		/**/
		/*	| OLCUC		  */
			| ONLCR		/**/
		/*	| OCRNL		  */
		/*	| ONOCR		  */
		/*	| ONLRET	  */
		/*	| OFILL		  */
		/*	| OFDEL		  */
		/*	no delay	  */
			;
    term->c_cflag =	0x0
			| CS8		/* CSIZE */
		/*	| CSTOPB	  */
			| CREAD		/**//* do not clear */
		/*	| PARENB	  */
		/*	| PARODD	  */
			| HUPCL		/**//* do not clear */
			| CLOCAL	/**//* do not clear */
			| CRTSCTS	/**//* do not clear */
			;
    term->c_lflag =	0x0
			| ISIG		/**/
		/*	| ICANON	  *//* do not stand */
		/*	| XCASE		  */
		/*	| ECHO		  *//* do not stand */
		/*	| ECHOE		  */
		/*	| ECHOK		  */
		/*	| ECHONL	  */
		/*	| NOFLSH	  */
		/*	| TOSTOP	  */
			| ECHOCTL	/**/
		/*	| ECHOPRT	  */
		/*	| ECHOKE	  */
		/*	| FLUSHO	  */
			| NOFLSH	/**/
		/*	| PENDIN	  */
			| IEXTEN	/**/
			;

    cfsetospeed(term, B9600);
    cfsetispeed(term, 0    );

#ifndef CTRL
# define CTRL(X) (X - '@')
#endif
    term->c_cc[VINTR]  = CTRL('C');
    term->c_cc[VQUIT]  = CTRL('\\');
    term->c_cc[VERASE] = '\177';
    term->c_cc[VWERASE]= CTRL('W');
    term->c_cc[VKILL]  = CTRL('U');
    term->c_cc[VREPRINT]=CTRL('R');
    term->c_cc[VEOF]   = CTRL('D');
    term->c_cc[VEOL]   = '\r';
    term->c_cc[VEOL2]  = '\0';
    term->c_cc[VSUSP]  = CTRL('Z');
    term->c_cc[VSTOP]  = CTRL('S');
    term->c_cc[VSTART] = CTRL('Q');
    term->c_cc[VDISCARD]=CTRL('O');
    term->c_cc[VLNEXT] = CTRL('V');

    term->c_cc[VMIN]  = 1;
    term->c_cc[VTIME] = 0;

    while(*cpar)
    {

    	switch(*cpar++)
	{
	case 'B':
	case 'b':
	    for(tp = baudt; tp->name; tp++)
	        if (!strncmp(tp->name, cpar, strlen(tp->name)))
			break;

	    if (!tp->name)
	    {
	    	writemsg(SYSTM, "%s unknown baud rate %s.\n", tty, cpar);
		while('0' <= *cpar && *cpar <= '9')
		    cpar++;
		continue;
	    }
	    cpar += strlen(tp->name);

	    cfsetospeed(term, tp->val);
	    cfsetispeed(term, tp->val);

	    ret = tp->name;

	    if (trace & TRACE_FILE)
		writemsg(TRACE, "baud: %s \n", tp->name);
	    break;

	case 'C':
	case 'c':
	    for(tp = csizet; tp->name; tp++)
	        if (!strncmp(tp->name, cpar, strlen(tp->name)))
			break;

	    if (!tp->name)
	    {
	    	writemsg(SYSTM, "%s unknown char size %s.\n", tty, cpar);
		while('0' <= *cpar && *cpar <= '9')
		    cpar++;
		continue;
	    }

	    cpar += strlen(tp->name);
	    term->c_cflag &= ~ CSIZE;
	    term->c_cflag |= tp->val;

	    if (trace & TRACE_FILE)
		writemsg(TRACE, "csize: %s\n", tp->name);
	    break;

	case 'S':
	case 's':
	    for(tp = cstopt; tp->name; tp++)
	        if (!strncmp(tp->name, cpar, strlen(tp->name)))
			break;

	    if (!tp->name)
	    {
	    	writemsg(SYSTM, "%s unknown stop bit length %s.\n", tty, cpar);
		while('0' <= *cpar && *cpar <= '9')
		    cpar++;
		continue;
	    }

	    cpar += strlen(tp->name);
	    term->c_cflag &= ~ CSTOPB;
	    term->c_cflag |= tp->val;

	    if (trace & TRACE_FILE)
		writemsg(TRACE, "cstop: %s\n", tp->name);
	    break;

	case 'P':
	case 'p':
	    for(tp = cpart; tp->name; tp++)
	        if (!strncmp(tp->name, cpar, strlen(tp->name)))
			break;

	    if (!tp->name)
	    {
	    	writemsg(SYSTM, "%s unknown parity %s.\n", tty, cpar);
		if (*cpar)
		    cpar++;
		continue;
	    }

	    cpar += strlen(tp->name);
	    term->c_cflag &= (~PARENB & ~PARODD);
	    term->c_cflag |= tp->val;

	    if (trace & TRACE_FILE)
		writemsg(TRACE, "cparity: %s\n", tp->name);
	    break;

	case 'F':
	case 'f':
	    for(tp = cflow; tp->name; tp++)
	        if (!strncmp(tp->name, cpar, strlen(tp->name)))
			break;

	    if (!tp->name)
	    {
	    	writemsg(SYSTM, "%s unknown flow control %s.\n", tty, cpar);
		if (*cpar)
		    cpar++;
		continue;
	    }

	    cpar += strlen(tp->name);
	    term->c_iflag &= ~(IXON | IXOFF | IXANY);
	    term->c_cflag &= ~CRTSCTS;

	    switch(tp->val)
	    {
	    case 2:	/* hardware flow */
		term->c_cflag |= CRTSCTS;
		tp->name = "hardware";
		break;
	    case 3:
		term->c_iflag |= (IXON | IXOFF | IXANY);
		tp->name = "xon/xoff";
		break;
	    default:
		tp->name = "none";
	    }

	    if (trace & TRACE_FILE)
		writemsg(TRACE, "cflow: %s\n", tp->name);
	    break;

	default:
	    writemsg(SYSTM, "%s unknown terminal parameter %s.\n", tty, cpar - 1);
	}

    }

    return ret;
}


/*
 *	code to dump terminal settings
 */

static char *
prchr(c)
    char c;
{
    static char cbuf[10];

    if (c < ' ')
    {
	sprintf(cbuf, "Ctrl-%c", c + '@');
    	return cbuf;
    }

    if (c == ' ')
    	return "Space";

    if (c == '\177')
    	return "Del";

    return NUL;
}

void
dumptio(tio)
    struct termios *tio;
{ 
    struct name_val *tp;
    speed_t sp;

    writemsg(TRACE, "VMIN = %d  VTIME = %d\n",
				tio->c_cc[VMIN], tio->c_cc[VTIME]);

    sp = cfgetispeed(tio);
    for(tp = baudt; tp->val != sp && tp->name; tp++);
    if (tp->name)
    	writemsg(TRACE, "Input baud rate:  %s\n", tp->name);
    else
    	writemsg(SYSTM, "unsupported input baud rate.\n");

    sp = cfgetospeed(tio);
    for(tp = baudt; tp->val != sp && tp->name; tp++);
    if (tp->name)
    	writemsg(TRACE, "Output baud rate: %s\n", tp->name);
    else
    	writemsg(SYSTM, "unsupported output baud rate.\n");

    writemsg(TRACE, "-----------------\n");

    writemsg(TRACE, "if: IGNBRK %s\n", tio->c_iflag & IGNBRK ? "on": "off");
    writemsg(TRACE, "if: BRKINT %s\n", tio->c_iflag & BRKINT ? "on": "off");
    writemsg(TRACE, "if: IGNPAR %s\n", tio->c_iflag & IGNPAR ? "on": "off");
    writemsg(TRACE, "if: PARMRK %s\n", tio->c_iflag & PARMRK ? "on": "off");
    writemsg(TRACE, "if: INPCK  %s\n", tio->c_iflag & INPCK  ? "on": "off");
    writemsg(TRACE, "if: ISTRIP %s\n", tio->c_iflag & ISTRIP ? "on": "off");
    writemsg(TRACE, "if: INLCR  %s\n", tio->c_iflag & INLCR  ? "on": "off");
    writemsg(TRACE, "if: IGNCR  %s\n", tio->c_iflag & IGNCR  ? "on": "off");
    writemsg(TRACE, "if: ICRNL  %s\n", tio->c_iflag & ICRNL  ? "on": "off");
#ifdef IUCLC
    writemsg(TRACE, "if: IUCLC  %s\n", tio->c_iflag & IUCLC  ? "on": "off");
#endif
    writemsg(TRACE, "if: IXON   %s\n", tio->c_iflag & IXON   ? "on": "off");
    writemsg(TRACE, "if: IXANY  %s\n", tio->c_iflag & IXANY  ? "on": "off");
    writemsg(TRACE, "if: IXOFF  %s\n", tio->c_iflag & IXOFF  ? "on": "off");
    writemsg(TRACE, "if: IMAXBEL %s\n",tio->c_iflag & IMAXBEL ? "on": "off");

    writemsg(TRACE, "-----------------\n");

    writemsg(TRACE, "of: OPOST  %s\n", tio->c_oflag & OPOST  ? "on": "off");
#ifdef OLCUC
    writemsg(TRACE, "of: OLCUC  %s\n", tio->c_oflag & OLCUC  ? "on": "off");
#endif
    writemsg(TRACE, "of: ONLCR  %s\n", tio->c_oflag & ONLCR  ? "on": "off");
#ifdef OCRNL
    writemsg(TRACE, "of: OCRNL  %s\n", tio->c_oflag & OCRNL  ? "on": "off");
#endif
#ifdef ONOCR
    writemsg(TRACE, "of: ONOCR  %s\n", tio->c_oflag & ONOCR  ? "on": "off");
#endif
#ifdef ONLRET
    writemsg(TRACE, "of: ONLRET %s\n", tio->c_oflag & ONLRET ? "on": "off");
#endif
#ifdef OFILL
    writemsg(TRACE, "of: OFILL  %s\n", tio->c_oflag & OFILL  ? "on": "off");
#endif
#ifdef OFDEL
    writemsg(TRACE, "of: OFDEL  %s\n", tio->c_oflag & OFDEL  ? "on": "off");
#endif

    writemsg(TRACE, "-----------------\n");

    writemsg(TRACE, "cf: CS8    %s\n", tio->c_cflag & CS8    ? "on": "off");
    writemsg(TRACE, "cf: CSTOPB %s\n", tio->c_cflag & CSTOPB ? "on": "off");
    writemsg(TRACE, "cf: CREAD  %s\n", tio->c_cflag & CREAD  ? "on": "off");
    writemsg(TRACE, "cf: PARENB %s\n", tio->c_cflag & PARENB ? "on": "off");
    writemsg(TRACE, "cf: PARODD %s\n", tio->c_cflag & PARODD ? "on": "off");
    writemsg(TRACE, "cf: HUPCL  %s\n", tio->c_cflag & HUPCL  ? "on": "off");
    writemsg(TRACE, "cf: CLOCAL %s\n", tio->c_cflag & CLOCAL ? "on": "off");
    writemsg(TRACE, "cf: CRTSCTS %s\n", tio->c_cflag & CRTSCTS ? "on": "off");

    writemsg(TRACE, "-----------------\n");

    writemsg(TRACE, "lf: ISIG   %s\n", tio->c_lflag & ISIG   ? "on": "off");
    writemsg(TRACE, "lf: ICANON %s\n", tio->c_lflag & ICANON ? "on": "off");
#ifdef XCASE
    writemsg(TRACE, "lf: XCASE  %s\n", tio->c_lflag & XCASE  ? "on": "off");
#endif
    writemsg(TRACE, "lf: ECHO   %s\n", tio->c_lflag & ECHO   ? "on": "off");
    writemsg(TRACE, "lf: ECHOE  %s\n", tio->c_lflag & ECHOE  ? "on": "off");
    writemsg(TRACE, "lf: ECHOK  %s\n", tio->c_lflag & ECHOK  ? "on": "off");
    writemsg(TRACE, "lf: ECHONL %s\n", tio->c_lflag & ECHONL ? "on": "off");
    writemsg(TRACE, "lf: NOFLSH %s\n", tio->c_lflag & NOFLSH ? "on": "off");
    writemsg(TRACE, "lf: TOSTOP %s\n", tio->c_lflag & TOSTOP ? "on": "off");
    writemsg(TRACE, "lf: ECHOCTL %s\n", tio->c_lflag & ECHOCTL ? "on": "off");
    writemsg(TRACE, "lf: ECHOPRT %s\n", tio->c_lflag & ECHOPRT ? "on": "off");
    writemsg(TRACE, "lf: ECHOKE %s\n", tio->c_lflag & ECHOKE ? "on": "off");
    writemsg(TRACE, "lf: FLUSHO %s\n", tio->c_lflag & FLUSHO ? "on": "off");
    writemsg(TRACE, "lf: PENDIN %s\n", tio->c_lflag & PENDIN ? "on": "off");
    writemsg(TRACE, "lf: IEXTEN %s\n", tio->c_lflag & IEXTEN ? "on": "off");

    writemsg(TRACE, "-----------------\n");

    writemsg(TRACE, "VINTR   = %s\n", prchr(tio->c_cc[VINTR   ]));
    writemsg(TRACE, "VQUIT   = %s\n", prchr(tio->c_cc[VQUIT   ]));
    writemsg(TRACE, "VERASE  = %s\n", prchr(tio->c_cc[VERASE  ]));
    writemsg(TRACE, "VWERASE = %s\n", prchr(tio->c_cc[VWERASE ]));
    writemsg(TRACE, "VKILL   = %s\n", prchr(tio->c_cc[VKILL   ]));
    writemsg(TRACE, "VREPRINT= %s\n", prchr(tio->c_cc[VREPRINT]));
    writemsg(TRACE, "VEOF    = %s\n", prchr(tio->c_cc[VEOF    ]));
    writemsg(TRACE, "VEOL    = %s\n", prchr(tio->c_cc[VEOL    ]));
    writemsg(TRACE, "VEOL2   = %s\n", prchr(tio->c_cc[VEOL2   ]));
    writemsg(TRACE, "VSUSP   = %s\n", prchr(tio->c_cc[VSUSP   ]));
    writemsg(TRACE, "VSTOP   = %s\n", prchr(tio->c_cc[VSTOP   ]));
    writemsg(TRACE, "VSTART  = %s\n", prchr(tio->c_cc[VSTART  ]));
    writemsg(TRACE, "VDISCARD= %s\n", prchr(tio->c_cc[VDISCARD]));
    writemsg(TRACE, "VLNEXT  = %s\n", prchr(tio->c_cc[VLNEXT  ]));
}


